from typing import Any, ClassVar, Optional, Type

from dictdaora import DictDaora
from jsondaora import as_typed_dict_field


BUILTIN_TYPE = type


class Header(DictDaora):
    type: ClassVar[Type[Any]]
    http_name: ClassVar[Optional[str]]

    def __init__(self, value: Any):
        value = as_typed_dict_field(value, 'value', type(self).type)
        super().__init__(value=value)

    def __init_subclass__(
        cls, type: Type[Any], http_name: Optional[str] = None
    ) -> None:
        cls.type = type
        cls.http_name = http_name


class LocationHeader(Header, type=str, http_name='location'):
    ...



from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Type

from .header import Header
from .route.controller_input import ControllerInput


@dataclass
class Request:
    path_pattern: str
    resolved_path: str
    controller_input_cls: Type[ControllerInput]
    path_args: Optional[Dict[str, Any]] = None
    query_dict: Optional[Dict[str, Any]] = None
    headers: Optional[Dict[str, Header]] = None
    body: Any = None
    ctx: Dict[str, Any] = field(default_factory=dict)


def make_controller_input_from_request(request: Request) -> ControllerInput:
    kwargs = {}

    if request.path_args:
        kwargs.update(request.path_args)
    if request.query_dict:
        kwargs.update(request.query_dict)
    if request.headers:
        kwargs.update(request.headers)
    if request.body:
        kwargs['body'] = request.body
    if request.controller_input_cls.__annotations_info__.has_kwargs:
        kwargs.update(request.ctx)
        kwargs['request'] = request

    return request.controller_input_cls(**kwargs)



import functools
import itertools
from asyncio import iscoroutine
from dataclasses import is_dataclass
from http import HTTPStatus
from json import JSONDecodeError
from typing import (  # type: ignore
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Type,
    Union,
    _GenericAlias,
)

import orjson
from jsondaora import dataclass_asjson, typed_dict_asjson
from jsondaora.deserializers import deserialize_field
from jsondaora.exceptions import DeserializationError

from ..asgi.base import ASGICallableResults, ASGIHeaders, ASGIResponse
from ..asgi.request import AsgiRequest
from ..asgi.responses import (
    make_html_response,
    make_json_response,
    make_no_content_response,
    make_not_found_response,
    make_see_other_response,
    make_text_response,
    make_yaml_response,
)
from ..asgi.router import Controller, Route
from ..bodies import GZipFactory
from ..content import ContentType
from ..exceptions import BadRequestError, InvalidReturnError
from ..header import Header
from ..method import MethodType
from ..middlewares import Middlewares
from ..request import Request, make_controller_input_from_request
from ..responses import Response
from .controller_input import controller_input


RESPONSES_MAP: Dict[
    Union[ContentType, HTTPStatus], Union[Callable[..., ASGIResponse]],
] = {
    ContentType.APPLICATION_JSON: make_json_response,
    ContentType.TEXT_PLAIN: make_text_response,
    ContentType.TEXT_HTML: make_html_response,
    ContentType.APPLICATION_YAML: make_yaml_response,
    HTTPStatus.NOT_FOUND: make_not_found_response,
    HTTPStatus.NO_CONTENT: make_no_content_response,
    HTTPStatus.SEE_OTHER: make_see_other_response,
}


def make_route(
    path_pattern: str,
    method: MethodType,
    controller: Union[Callable[..., Any], Controller],
    has_content_length: bool = True,
    route_middlewares: Optional[Union[Middlewares]] = None,
    options: bool = False,
) -> Route:
    ControllerInput = controller_input(controller, path_pattern)
    annotations_info = ControllerInput.__annotations_info__
    annotations_path_args = ControllerInput.__annotations_path_args__
    annotations_query_dict = ControllerInput.__annotations_query_dict__
    annotations_headers = ControllerInput.__annotations_headers__
    body_type = ControllerInput.__annotations_body__.get('body')
    return_type = controller.__annotations__.get('return')

    def parse_asgi_input(
        asgi_request: AsgiRequest, middlewares: Optional[Middlewares] = None,
    ) -> Request:
        request = Request(
            asgi_request.path_pattern,
            asgi_request.resolved_path,
            ControllerInput,
        )

        if annotations_info.has_input:
            if annotations_info.has_path_args:
                request.path_args = {
                    name: deserialize_field(
                        name, type_, asgi_request.path_args.get(name)
                    )
                    for name, type_ in annotations_path_args.items()
                }

            if annotations_info.has_query_dict:
                request.query_dict = {
                    name: deserialize_field(
                        name,
                        type_,
                        parse_query_args(
                            name, type_, asgi_request.query_dict.get(name)
                        ),
                    )
                    for name, type_ in annotations_query_dict.items()
                    if name != 'body'
                }

            if annotations_info.has_headers:
                headers_map = ControllerInput.__headers_name_map__
                request.headers = {
                    name: annotations_headers[name](h_value.decode())  # type: ignore
                    for h_name, h_value in asgi_request.headers
                    if (name := headers_map.get(h_name.decode()))  # noqa
                    in annotations_headers
                }

            if annotations_info.has_body:
                if isinstance(body_type, type) and issubclass(
                    body_type, GZipFactory
                ):
                    request.body = body_type(value=asgi_request.body)

                elif isinstance(body_type, type) and issubclass(
                    body_type, str
                ):
                    request.body = asgi_request.body.decode()

                else:
                    request.body = make_json_request_body(
                        asgi_request.body, body_type
                    )
                    request.body = (
                        deserialize_field('body', body_type, request.body)
                        if request.body
                        else None
                    )

        return request

    async def build_asgi_output(
        request: Request,
        controller_output: Any,
        status: HTTPStatus = HTTPStatus.OK,
        headers: Optional[Sequence[Header]] = None,
        content_type: Optional[ContentType] = ContentType.APPLICATION_JSON,
        return_type_: Any = None,
        middlewares: Optional[Middlewares] = None,
    ) -> ASGICallableResults:
        while iscoroutine(controller_output):
            controller_output = await controller_output

        if return_type_ is None and return_type:
            return_type_ = return_type

        if middlewares:
            if not isinstance(controller_output, Response):
                controller_output = Response(
                    body=controller_output,
                    status=status,
                    headers=headers,
                    content_type=content_type,
                )

            for middleware in middlewares.post_execution:
                middleware(request, controller_output)

        if isinstance(controller_output, Response):
            return build_asgi_output(
                request,
                controller_output['body'],
                controller_output['status'],
                controller_output['headers'],
                controller_output['content_type'],
                controller_output.__annotations__.get('body'),
            )

        elif isinstance(controller_output, dict):
            if return_type_:
                body = typed_dict_asjson(controller_output, return_type_)
            else:
                body = orjson.dumps(controller_output)

        elif (
            is_dataclass(controller_output)
            or isinstance(controller_output, tuple)
            or isinstance(controller_output, list)
        ):
            body = dataclass_asjson(controller_output)

        elif (
            isinstance(controller_output, str)
            or isinstance(controller_output, int)
            or isinstance(controller_output, bool)
            or isinstance(controller_output, float)
            or isinstance(controller_output, bytes)
        ):
            if content_type == ContentType.APPLICATION_JSON:
                if isinstance(controller_output, bytes):
                    body = controller_output
                else:
                    body = orjson.dumps(controller_output)
            else:
                body = str(controller_output).encode()

        elif controller_output is None:
            content_length = 0 if has_content_length else None

            if content_type is None and status in RESPONSES_MAP:
                if headers:
                    return (RESPONSES_MAP[status](make_asgi_headers(headers)),)

                return (RESPONSES_MAP[status](headers),)

            if headers:
                return RESPONSES_MAP[content_type](  # type: ignore
                    content_length, headers=make_asgi_headers(headers)
                )

            return RESPONSES_MAP[content_type](content_length)  # type: ignore

        else:
            raise InvalidReturnError(controller_output, controller)

        content_length = len(body) if has_content_length else None

        return (
            RESPONSES_MAP[content_type](  # type: ignore
                content_length, status, make_asgi_headers(headers)
            ),
            body,
        )

    class WrappedController(Controller):
        @functools.wraps(controller)
        async def __call__(
            self, asgi_request: AsgiRequest,
        ) -> Union[Awaitable[ASGICallableResults], ASGICallableResults]:
            try:
                request = parse_asgi_input(asgi_request,)
                if self.middlewares:
                    for middleware in self.middlewares.pre_execution:
                        middleware(request)

                controller_output = controller(
                    **make_controller_input_from_request(request)
                )

                return await build_asgi_output(
                    request, controller_output, middlewares=self.middlewares,
                )

            except BadRequestError as error:
                if self.logger:
                    error_attrs = ' '.join(
                        [f'{k}={v}' for k, v in error.dict.items()]
                    )
                    self.logger.warning(f"BAD REQUEST ERROR {error_attrs}")

                return send_bad_request_response(
                    error.dict, has_content_length, error.headers
                )

            except DeserializationError as error:
                error_dict = {
                    'name': 'field-parsing',
                    'message': error.args[0],
                }
                return send_bad_request_response(
                    error_dict, has_content_length
                )

    wrapped_controller = WrappedController()
    route = Route(
        path_pattern,
        method,
        wrapped_controller,
        annotations_info.has_path_args,
        annotations_info.has_query_dict,
        annotations_info.has_headers,
        annotations_info.has_body,
        has_options=options,
    )
    routes = [route]

    if isinstance(controller, Controller) and hasattr(controller, 'routes'):
        routes = controller.routes + routes

    wrapped_controller.routes = routes
    wrapped_controller.middlewares = route_middlewares

    return route


def make_json_request_body(body: bytes, body_type: Optional[Type[Any]]) -> Any:
    try:
        return orjson.loads(body)
    except JSONDecodeError:
        schema = (
            getattr(body_type, '__annotations__', {}) if body_type else None
        )
        schema = {k: t.__name__ for k, t in schema.items()}
        raise BadRequestError(name='invalid-body', info={'schema': schema})


def send_bad_request_response(
    error_dict: Dict[str, Any],
    has_content_length: bool,
    headers: Optional[Sequence[Header]] = None,
    content_type: ContentType = ContentType.APPLICATION_JSON,
) -> ASGICallableResults:
    body = orjson.dumps({'error': error_dict})
    content_length = len(body) if has_content_length else None

    return (
        RESPONSES_MAP[content_type](
            content_length, HTTPStatus.BAD_REQUEST, make_asgi_headers(headers)
        ),
        body,
    )


def make_asgi_headers(
    headers: Optional[Sequence[Header]],
) -> Optional[ASGIHeaders]:
    if not headers:
        return None

    return [
        (
            header.http_name.encode(),  # type: ignore
            str(header.value).encode()
            if not isinstance(header.value, bytes)
            else header.value,
        )
        for header in headers
    ]


def parse_query_args(
    name: str, type_: Type[Any], value: Optional[List[str]]
) -> Any:
    if value is not None:
        if isinstance(type_, _GenericAlias) and type_._name in (
            'List',
            'Tuple',
            'Set',
            'Deque',
        ):
            value = list(itertools.chain(*[v.split(',') for v in value]))

        elif len(value) > 1:
            raise BadRequestError(
                name='invalid-query',
                info={'name': name, 'type': type_, 'value': value},
            )
        else:
            value = value[0]  # type: ignore

        return value



import asyncio

from apidaora import appdaora, route


# async tasks can potentially block yours application, use with care
# it use is recommended just for small tasks or non-blocking operations
@route.background('/hello')
async def hello_task(name: str) -> str:
    await asyncio.sleep(1)
    return f'Hello {name}!'


app = appdaora(hello_task)



from enum import Enum


class ContentType(Enum):
    APPLICATION_JSON = 'application/json'
    APPLICATION_YAML = 'application/x-yaml'
    TEXT_HTML = 'text/html; charset=utf-8'
    TEXT_PLAIN = 'text/plain'



from dataclasses import dataclass
from typing import Any, Dict, List, Tuple


@dataclass
class AsgiRequest:
    path_pattern: str
    resolved_path: str
    path_args: Dict[str, Any]
    query_dict: Dict[str, Any]
    headers: List[Tuple[bytes, bytes]]
    body: bytes



from typing import Any

from apidaora import Middlewares, Request, Response, appdaora, route, text


def request_extra_args(request: Request) -> None:
    request.ctx['extra'] = 'You'


def response_extra_args(request: Request, response: Response) -> None:
    if request.ctx and response.ctx:
        response.body = response.body.replace(
            'You', f"{request.ctx['extra']} and {response.ctx['name']}"
        )


@route.get(
    '/middlewares-ctx',
    middlewares=Middlewares(
        pre_execution=request_extra_args, post_execution=response_extra_args,
    ),
)
async def extra_args_controller(name: str, **kwargs: Any) -> Response:
    return text(hello(kwargs['extra']), name=name)


def hello(name: str) -> str:
    return f'Hello {name}!'


app = appdaora(extra_args_controller)



import asyncio
import dataclasses
import datetime
import hashlib
import logging
import uuid
from concurrent.futures import ThreadPoolExecutor
from enum import Enum
from functools import partial
from http import HTTPStatus
from typing import (
    Any,
    Awaitable,
    Callable,
    Coroutine,
    Dict,
    List,
    Optional,
    Type,
    TypedDict,
)

import orjson
from jsondaora import as_typed_dict, jsondaora, typed_dict_asjson

from ..asgi.router import Controller
from ..exceptions import BadRequestError, InvalidTasksRepositoryError
from ..header import Header, LocationHeader
from ..method import MethodType
from ..middlewares import Middlewares
from ..responses import Response, json, see_other
from ..route.factory import make_route


logger = logging.getLogger(__name__)


try:
    import aioredis
except Exception:
    aioredis = None


class TaskStatusType(Enum):
    RUNNING = 'running'
    FINISHED = 'finished'
    ERROR = 'error'


@jsondaora
class TaskInfo(TypedDict):
    task_id: str
    start_time: str
    status: str
    signature: str
    args_signature: Optional[str]


@dataclasses.dataclass
class BackgroundTask:
    create: Controller
    get_results: Controller


def make_background_task(
    controller: Callable[..., Any],
    path_pattern: str,
    max_workers: int = 10,
    tasks_repository_uri: Optional[str] = None,
    lock: bool = False,
    lock_args: bool = False,
    middlewares: Optional[Middlewares] = None,
    options: bool = False,
) -> BackgroundTask:
    if asyncio.iscoroutinefunction(controller):
        logger.warning(
            'Async tasks can potentially block your application, use with care. '
            'It use is recommended just for small tasks or non-blocking operations.'
        )

    annotations = getattr(controller, '__annotations__', {})
    result_annotation = annotations.get('return', str)
    signature = hashlib.md5(
        '-'.join(
            [controller.__name__]
            + [
                '-'.join((name, type_.__name__))
                for name, type_ in annotations.items()
            ]
        ).encode()
    ).hexdigest()[:12]
    tasks_repository_builder = get_tasks_repository_builder(
        tasks_repository_uri, signature
    )

    @jsondaora
    class FinishedTaskInfo(TaskInfo):
        end_time: str
        result: result_annotation  # type: ignore

    create_task = make_create_task(
        controller,
        tasks_repository_builder,
        FinishedTaskInfo,
        max_workers,
        lock,
        lock_args,
        signature,
    )
    get_task_results = make_get_task_results(
        tasks_repository_builder, FinishedTaskInfo
    )

    create_task.__annotations__ = {
        name: type_ for name, type_ in annotations.items() if name != 'return'
    }

    if 'kwargs' not in create_task.__annotations__:
        create_task.__annotations__['kwargs'] = Any

    return BackgroundTask(
        make_route(
            path_pattern,
            MethodType.POST,
            create_task,
            route_middlewares=middlewares,
            options=options,
        ).controller,
        make_route(
            path_pattern,
            MethodType.GET,
            get_task_results,
            route_middlewares=middlewares,
            options=options,
        ).controller,
    )


def get_iso_time() -> str:
    return (
        datetime.datetime.now()
        .replace(microsecond=0, tzinfo=datetime.timezone.utc)
        .isoformat()
    )


def get_tasks_repository_builder(uri: Optional[str], signature: str) -> Any:
    if uri is None:

        async def builder() -> SimpleTasksRepository:
            return SimpleTasksRepository(signature, TASKS_DB)

        return builder

    elif isinstance(uri, str) and uri.startswith('redis://'):
        if aioredis is None:
            raise InvalidTasksRepositoryError("'aioredis' package not found!")

        return partial(
            get_redis_tasks_repository, signature=signature, uri=uri,
        )

    raise InvalidTasksRepositoryError(uri)


def make_create_task(
    controller: Callable[..., Any],
    tasks_repository_builder: Callable[[], Awaitable['BaseTasksRepository']],
    finished_task_info_cls: Any,
    max_workers: int,
    lock: bool,
    lock_args: bool,
    signature: str,
) -> Callable[..., Coroutine[Any, Any, Response]]:
    executor = ThreadPoolExecutor(max_workers)

    async def create_task(*args: Any, **kwargs: Any) -> Response:
        task_id = str(uuid.uuid4())
        args_signature = None
        resolved_path = kwargs['request'].resolved_path
        controller_annotations = getattr(controller, '__annotations__', {})

        if (
            'kwargs' not in controller_annotations
            and 'request' not in controller_annotations
        ):
            kwargs.pop('request')

        if lock_args:
            args_signature = hashlib.md5(
                '-'.join(
                    [str(arg) for arg in args]
                    + [
                        '-'.join((key, str(value)))
                        for key, value in kwargs.items()
                    ]
                ).encode()
            ).hexdigest()[:12]

        tasks_repository = await tasks_repository_builder()
        lock_response = await make_lock_response(
            tasks_repository, task_id, lock, args_signature, resolved_path
        )

        if lock_response:
            await tasks_repository.close()
            return lock_response

        if lock:
            await tasks_repository.set_locked_task_id(task_id)
        elif args_signature:
            await tasks_repository.set_locked_task_id(task_id, args_signature)

        if asyncio.iscoroutinefunction(controller):
            wrapper_async = make_controller_wrapper_async(
                tasks_repository_builder,
                task_id,
                lock,
                args_signature,
                finished_task_info_cls,
                controller,
                *args,
                **kwargs,
            )
            task_ = asyncio.create_task(wrapper_async())
            task_.add_done_callback(lambda f: f.result())

        else:
            wrapper = make_controller_wrapper(
                tasks_repository_builder,
                task_id,
                lock,
                args_signature,
                finished_task_info_cls,
                controller,
                *args,
                **kwargs,
            )
            future = executor.submit(wrapper)
            future.add_done_callback(lambda f: f.result())

        start_time = get_iso_time()

        task = TaskInfo(
            task_id=str(task_id),
            start_time=start_time,
            status=TaskStatusType.RUNNING.value,
            signature=signature,
            args_signature=args_signature,
        )
        await tasks_repository.set(task, task_id, finished_task_info_cls)
        await tasks_repository.close()

        return json(task, status=HTTPStatus.ACCEPTED)

    return create_task


def make_controller_wrapper_async(
    tasks_repository_builder: Callable[[], Awaitable['BaseTasksRepository']],
    task_id: str,
    lock: bool,
    args_signature: Optional[str],
    finished_task_info_cls: Any,
    controller: Any,
    *args: Any,
    **kwargs: Any,
) -> Callable[..., Coroutine[Any, Any, TaskInfo]]:
    async def wrapper() -> Any:
        tasks_repository = await tasks_repository_builder()

        try:
            result = await controller(*args, **kwargs)
            status = TaskStatusType.FINISHED.value
        except Exception as error:
            logger.exception(
                f'signature={tasks_repository.signature}; '
                f'args-signature={args_signature};'
                f'task-id={task_id};'
            )
            result = {
                'error': {
                    'name': type(error).__name__,
                    'args': [str(a) for a in error.args],
                }
            }
            status = TaskStatusType.ERROR.value

        task = await tasks_repository.get(task_id, finished_task_info_cls)
        finished_task = finished_task_info_cls(
            end_time=get_iso_time(),
            result=result,
            status=status,
            task_id=task['task_id'],
            start_time=task['start_time'],
            signature=task['signature'],
            args_signature=task['args_signature'],
        )
        await tasks_repository.set(
            finished_task, task_id, finished_task_info_cls
        )

        if lock:
            await tasks_repository.delete_locked_task_id()
        elif args_signature:
            await tasks_repository.delete_locked_task_id(args_signature)

        await tasks_repository.close()

    return wrapper


def make_controller_wrapper(
    tasks_repository_builder: Callable[[], Awaitable['BaseTasksRepository']],
    task_id: str,
    lock: bool,
    args_signature: Optional[str],
    finished_task_info_cls: Any,
    controller: Any,
    *args: Any,
    **kwargs: Any,
) -> Callable[..., Any]:
    def wrapper() -> Any:
        policy = asyncio.get_event_loop_policy()
        loop = policy.new_event_loop()
        tasks_repository = loop.run_until_complete(tasks_repository_builder())

        try:
            result = controller(*args, **kwargs)
            status = TaskStatusType.FINISHED.value
        except Exception as error:
            logger.exception(
                f'signature={tasks_repository.signature}; '
                f'args-signature={args_signature};'
                f'task-id={task_id};'
            )
            result = {
                'error': {
                    'name': type(error).__name__,
                    'args': [str(a) for a in error.args],
                }
            }
            status = TaskStatusType.ERROR.value

        task = loop.run_until_complete(
            tasks_repository.get(task_id, finished_task_info_cls)
        )
        finished_task = finished_task_info_cls(
            end_time=get_iso_time(),
            result=result,
            status=status,
            task_id=task['task_id'],
            start_time=task['start_time'],
            signature=task['signature'],
            args_signature=task['args_signature'],
        )

        loop.run_until_complete(
            tasks_repository.set(
                finished_task, task_id, finished_task_info_cls,
            )
        )

        if lock:
            loop.run_until_complete(tasks_repository.delete_locked_task_id())
        elif args_signature:
            loop.run_until_complete(
                tasks_repository.delete_locked_task_id(args_signature)
            )

        loop.run_until_complete(tasks_repository.close())

    return wrapper


def make_get_task_results(
    tasks_repository_builder: Callable[[], Awaitable['BaseTasksRepository']],
    finished_task_info_cls: Type[Any],
) -> Callable[..., Coroutine[Any, Any, TaskInfo]]:
    async def get_task_results(
        task_id: str, **kwargs: Any
    ) -> finished_task_info_cls:  # type: ignore
        tasks_repository = await tasks_repository_builder()

        try:
            task = await tasks_repository.get(task_id, finished_task_info_cls)
            return task  # type: ignore
        except KeyError:
            raise BadRequestError(
                name='invalid-task-id', info={'task_id': task_id}
            )
        except ValueError as error:
            if error.args == ('badly formed hexadecimal UUID string',):
                raise BadRequestError(
                    name='invalid-task-id', info={'task_id': task_id}
                )

            raise error from None
        finally:
            await tasks_repository.close()

    return get_task_results


async def make_lock_response(
    tasks_repository: 'BaseTasksRepository',
    task_id: str,
    lock: bool,
    args_signature: Optional[str],
    resolved_path: str,
) -> Optional[Response]:
    headers: List[Header] = []
    locked_task_id = None
    lock_type = None

    if lock:
        locked_task_id = await tasks_repository.get_locked_task_id()
        lock_type = LockType.SIGNATURE

    elif args_signature:
        locked_task_id = await tasks_repository.get_locked_task_id(
            args_signature
        )
        lock_type = LockType.ARGS_SIGNATURE

    if locked_task_id and lock_type:
        headers.append(
            LocationHeader(f'{resolved_path}?task_id={locked_task_id}')
        )
        headers.append(LockHeader(lock_type.value))
        return see_other(headers=headers)

    return None


class LockHeader(Header, type=str, http_name='x-apidaora-lock'):
    ...


class LockType(Enum):
    SIGNATURE = 'signature'
    ARGS_SIGNATURE = 'args-signature'


@dataclasses.dataclass
class BaseTasksRepository:
    signature: str
    data_source: Any

    async def set(
        self, value: Any, task_id: str, task_cls: Type[Any] = TaskInfo,
    ) -> None:
        raise NotImplementedError()

    async def get(self, task_id: str, finished_task_cls: Type[Any]) -> Any:
        raise NotImplementedError()

    async def set_locked_task_id(
        self, task_id: str, args_signature: Optional[str] = None
    ) -> None:
        raise NotImplementedError()

    async def delete_locked_task_id(
        self, args_signature: Optional[str] = None
    ) -> None:
        raise NotImplementedError()

    async def close(self) -> None:
        ...

    async def get_locked_task_id(
        self, args_signature: Optional[str] = None
    ) -> Optional[str]:
        raise NotImplementedError()

    def build_lock_key(self, args_signature: Optional[str]) -> str:
        if args_signature:
            return f'apidaora:{self.signature}:{args_signature}'

        return f'apidaora:{self.signature}'

    def build_key(self, task_id: str) -> str:
        return f'apidaora:{self.signature}:{task_id}'


@dataclasses.dataclass
class SimpleTasksRepository(BaseTasksRepository):
    data_source: Dict[str, Any]

    async def set(
        self, value: Any, task_id: str, task_cls: Type[Any] = TaskInfo,
    ) -> None:
        self.data_source[self.build_key(task_id)] = value

    async def get(self, task_id: str, finished_task_cls: Type[Any]) -> Any:
        return self.data_source[self.build_key(task_id)]

    async def set_locked_task_id(
        self, task_id: str, args_signature: Optional[str] = None
    ) -> None:
        self.data_source[self.build_lock_key(args_signature)] = task_id

    async def get_locked_task_id(
        self, args_signature: Optional[str] = None
    ) -> Optional[str]:
        return self.data_source.get(self.build_lock_key(args_signature))

    async def delete_locked_task_id(
        self, args_signature: Optional[str] = None
    ) -> None:
        self.data_source.pop(self.build_lock_key(args_signature), None)


if aioredis is not None:

    @dataclasses.dataclass
    class RedisTasksRepository(BaseTasksRepository):
        data_source: aioredis.Redis

        async def set(
            self, value: Any, task_id: str, task_cls: Type[Any] = TaskInfo,
        ) -> None:
            await self.data_source.set(
                self.build_key(task_id), typed_dict_asjson(value, task_cls)
            )

        async def get(self, task_id: str, finished_task_cls: Type[Any]) -> Any:
            value = await self.data_source.get(self.build_key(task_id))

            if value:
                value = orjson.loads(value)

                if value['status'] == TaskStatusType.RUNNING.value:
                    return as_typed_dict(value, TaskInfo)

                if value['status'] == TaskStatusType.FINISHED.value or (
                    value['status'] == TaskStatusType.ERROR.value
                ):
                    return as_typed_dict(value, finished_task_cls)

            raise KeyError(self.build_key(task_id))

        async def set_locked_task_id(
            self, task_id: str, args_signature: Optional[str] = None
        ) -> None:
            await self.data_source.set(
                self.build_lock_key(args_signature), task_id
            )

        async def get_locked_task_id(
            self, args_signature: Optional[str] = None
        ) -> Optional[str]:
            return await self.data_source.get(  # type: ignore
                self.build_lock_key(args_signature)
            )

        async def delete_locked_task_id(
            self, args_signature: Optional[str] = None
        ) -> None:
            await self.data_source.delete(self.build_lock_key(args_signature))

        async def close(self) -> None:
            self.data_source.close()
            await self.data_source.wait_closed()

    async def get_redis_tasks_repository(
        signature: str, uri: str
    ) -> RedisTasksRepository:
        data_source = await aioredis.create_redis_pool(uri)
        return RedisTasksRepository(signature, data_source)


TASKS_DB: Dict[str, Any] = {}



from typing import Any, Dict, Type

from ..header import Header


def get_annotations_path_args(
    path_pattern: str, annotations: Dict[str, Type[Any]]
) -> Dict[str, Type[Any]]:
    return {
        name: type_
        for name, type_ in annotations.items()
        if f'{{{name}}}' in path_pattern
    }


def get_annotations_query_dict(
    path_args_annotations: Dict[str, Type[Any]],
    headers_annotations: Dict[str, Type[Any]],
    annotations: Dict[str, Type[Any]],
) -> Dict[str, Type[Any]]:
    return {
        name: type_
        for name, type_ in annotations.items()
        if name not in path_args_annotations
        and name not in headers_annotations
        and name != 'return'
        and name != 'args'
        and name != 'kwargs'
        and name != 'request'
    }


def get_annotations_headers(
    headers_name_map: Dict[str, str], annotations: Dict[str, Type[Any]]
) -> Dict[str, Type[Any]]:
    annotations_headers: Dict[str, Type[Header]] = {}

    for name, type_ in annotations.items():
        if isinstance(type_, type) and issubclass(type_, Header):
            if type_.http_name is None:
                http_name = (
                    ''
                    if name.startswith('x-') or name.startswith('x_')
                    else 'x-'
                )
                http_name += name.replace('_', '-')
                type_.http_name = http_name
            else:
                http_name = type_.http_name

            annotations_headers[name] = type_
            headers_name_map[http_name] = name

    return annotations_headers


def get_annotations_body(
    annotations: Dict[str, Type[Any]]
) -> Dict[str, Type[Any]]:
    body_type = annotations.get('body')

    if body_type:
        return {'body': body_type}

    return {}



import asyncio
from typing import Any, Awaitable, Callable, Dict
from urllib import parse

from ..exceptions import MethodNotFoundError, PathNotFoundError
from .base import ASGIApp, Receiver, Scope, Sender
from .request import AsgiRequest
from .responses import (
    send_method_not_allowed_response,
    send_not_found,
    send_response,
)
from .router import ResolvedRoute


def asgi_app(router: Callable[[str, str], ResolvedRoute]) -> ASGIApp:
    async def controller(
        scope: Scope, receive: Receiver, send: Sender
    ) -> None:
        try:
            resolved = router(scope['path'], scope['method'])

        except PathNotFoundError:
            await send_not_found(send)

        except MethodNotFoundError:
            await send_method_not_allowed_response(send)

        else:
            route = resolved.route

            if route.has_headers:
                headers = scope['headers']
            else:
                headers = []

            if route.has_query:
                query_dict = _get_query_dict(scope)
            else:
                query_dict = {}

            if route.has_body:
                body = await _read_body(receive)
            else:
                body = b''

            response_and_body = route.controller(
                AsgiRequest(
                    path_pattern=route.path_pattern,
                    resolved_path=resolved.path,
                    path_args=resolved.path_args,
                    query_dict=query_dict,
                    headers=headers,
                    body=body,
                )
            )

            while asyncio.iscoroutine(response_and_body):
                response_and_body = await response_and_body

            response, body = (
                (response_and_body[0], response_and_body[1])
                if len(response_and_body) > 1
                else (response_and_body[0], b'')
            )
            await send_response(send, response, body)

    return controller


def _get_query_dict(scope: Dict[str, Any]) -> Dict[str, Any]:
    qs = parse.parse_qs(scope['query_string'].decode())
    return qs


async def _read_body(receive: Callable[[], Awaitable[Dict[str, Any]]]) -> Any:
    body = b''
    more_body = True

    while more_body:
        message = await receive()
        body += message.get('body', b'')
        more_body = message.get('more_body', False)

    return body



from http import HTTPStatus
from typing import Any, Dict, Optional, Sequence

from dictdaora import DictDaora

from .content import ContentType
from .header import Header


class Response(DictDaora):
    status: HTTPStatus
    headers: Sequence[Header]
    content_type: Optional[ContentType]
    ctx: Optional[Dict[str, Any]]


def json(
    body: Any,
    status: HTTPStatus = HTTPStatus.OK,
    headers: Sequence[Header] = (),
    **kwargs: Any,
) -> Response:
    return Response(
        body=body,
        status=status,
        headers=headers,
        content_type=ContentType.APPLICATION_JSON,
        ctx=kwargs,
    )


def text(
    body: Any,
    status: HTTPStatus = HTTPStatus.OK,
    headers: Sequence[Header] = (),
    **kwargs: Any,
) -> Response:
    return Response(
        body=body,
        status=status,
        headers=headers,
        content_type=ContentType.TEXT_PLAIN,
        ctx=kwargs,
    )


def html(
    body: Any,
    status: HTTPStatus = HTTPStatus.OK,
    headers: Sequence[Header] = (),
    **kwargs: Any,
) -> Response:
    return Response(
        body=body,
        status=status,
        headers=headers,
        content_type=ContentType.TEXT_HTML,
        ctx=kwargs,
    )


def no_content(headers: Sequence[Header] = (), **kwargs: Any) -> Response:
    return Response(
        status=HTTPStatus.NO_CONTENT,
        headers=headers,
        content_type=None,
        body=None,
        ctx=kwargs,
    )


def see_other(headers: Sequence[Header] = (), **kwargs: Any) -> Response:
    return Response(
        status=HTTPStatus.SEE_OTHER,
        headers=headers,
        content_type=None,
        body=None,
        ctx=kwargs,
    )


def not_found(
    body: Optional[Any] = None,
    headers: Sequence[Header] = (),
    content_type: Optional[ContentType] = None,
    **kwargs: Any,
) -> Response:
    return Response(
        body=body,
        status=HTTPStatus.NOT_FOUND,
        headers=headers,
        content_type=content_type,
        ctx=kwargs,
    )



from apidaora import (
    BackgroundTaskMiddleware,
    Middlewares,
    Response,
    appdaora,
    route,
    text,
)


class HelloCounter:
    counter = 1

    @classmethod
    def count(cls) -> None:
        cls.counter += 1


@route.get('/background-tasks')
async def background_tasks_controller(name: str) -> Response:
    return text(
        f'Hello {name}!\n{name} are the #{HelloCounter.counter}!',
        background_tasks=HelloCounter.count,
    )


app = appdaora(
    background_tasks_controller,
    middlewares=Middlewares(post_execution=BackgroundTaskMiddleware()),
)



import asyncio
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from logging import Logger, getLogger
from typing import Any, Callable, Iterable, List, Optional, Union

from .header import Header
from .request import Request
from .responses import Response


@dataclass(init=False)
class Middlewares:
    pre_execution: List[Callable[[Request], None]]
    post_execution: List[Callable[[Request, Response], None]]

    def __init__(
        self,
        pre_execution: Optional[
            Union[List[Callable[[Request], None]], Callable[[Request], None]]
        ] = None,
        post_execution: Optional[
            Union[
                List[Callable[[Request, Response], None]],
                Callable[[Request, Response], None],
            ]
        ] = None,
    ):

        if pre_execution is None:
            pre_execution = []

        elif not isinstance(pre_execution, Iterable):
            pre_execution = [pre_execution]

        elif not isinstance(pre_execution, list):
            pre_execution = list(pre_execution)

        if post_execution is None:
            post_execution = []

        elif not isinstance(post_execution, Iterable):
            post_execution = [post_execution]

        elif not isinstance(post_execution, list):
            post_execution = list(post_execution)

        self.pre_execution = pre_execution
        self.post_execution = post_execution


class AllowOriginHeader(
    Header, type=str, http_name='access-control-allow-origin'
):
    ...


class ExposeHeadersHeader(
    Header, type=str, http_name='access-control-expose-headers'
):
    ...


class AllowHeadersHeader(
    Header, type=str, http_name='access-control-allow-headers'
):
    ...


class AllowMethodsHeader(
    Header, type=str, http_name='access-control-allow-methods'
):
    ...


class CorsMiddleware:
    def __init__(
        self,
        *,
        allow_origin: Optional[str] = None,
        expose_headers: Optional[str] = None,
        allow_headers: Optional[str] = None,
        allow_methods: Optional[str] = None,
        servers_all: Optional[str] = '*',
        logger: Logger = getLogger(__name__),
    ):
        if (
            not allow_origin
            and not expose_headers
            and not allow_headers
            and not allow_methods
        ):
            self.headers = [
                AllowOriginHeader(servers_all),
                ExposeHeadersHeader(servers_all),
                AllowHeadersHeader(servers_all),
                AllowMethodsHeader(servers_all),
            ]
        else:
            self.headers = []
            if allow_origin:
                self.headers.append(AllowOriginHeader(allow_origin))
            if expose_headers:
                self.headers.append(ExposeHeadersHeader(expose_headers))
            if allow_headers:
                self.headers.append(AllowHeadersHeader(allow_headers))
            if allow_methods:
                self.headers.append(AllowMethodsHeader(allow_methods))

        self.headers_tuple = tuple(self.headers)
        self.logger = logger

    def __call__(self, request: Request, response: Response) -> None:
        if isinstance(response.headers, list):
            response.headers.extend(self.headers)

        elif isinstance(response.headers, tuple):
            response.headers = response.headers + self.headers_tuple

        elif response.headers is None:
            response.headers = self.headers_tuple


@dataclass(init=False)
class BackgroundTaskMiddleware:
    executor: ThreadPoolExecutor
    logger: Logger

    def __init__(
        self, max_workers: int = 100, logger: Logger = getLogger(__name__),
    ):
        self.executor = ThreadPoolExecutor(max_workers=max_workers)
        self.logger = logger

    def __call__(self, request: Request, response: Response) -> None:
        task: Any
        response_tasks: Union[
            Iterable[Callable[[], None]], Callable[[], None]
        ] = []

        if response.ctx:
            response_tasks = response.ctx.get(
                'background_tasks', response_tasks
            )

        if response_tasks and not isinstance(response_tasks, Iterable):
            response_tasks = (response_tasks,)

        for task in response_tasks:
            if asyncio.iscoroutinefunction(task):
                task = task()

            elif not asyncio.iscoroutine(task):
                future = self.executor.submit(task)
                future.add_done_callback(self.done_callback)
                return

            task = asyncio.create_task(task)
            task.add_done_callback(self.done_callback)

    def done_callback(self, future: Any) -> None:
        try:
            future.result()
        except Exception:
            self.logger.exception('Background Task Error')



from typing import TypedDict

from jsondaora import jsondaora, typed_dict_asjson

from apidaora.core import JSON_RESPONSE, appdaora_core, route


@jsondaora
class You(TypedDict):
    name: str
    location: str


@route.get('/hello/{name}', query=True)
async def hello_controller(request):  # type: ignore
    you = You(
        name=request.path_args['name'],
        location=request.query_dict['location'][0],
    )
    body = typed_dict_asjson(you, You)
    return JSON_RESPONSE, body


app = appdaora_core(hello_controller)
